from string import ascii_letters

import pandas as pd
from hypothesis import given
from hypothesis.strategies import lists, text

from trrex import make


@given(lists(text(alphabet=ascii_letters, min_size=1), min_size=1))
def test_extract(lst):
    pattern = make(lst, prefix=r"\b(", suffix=r")\b")
    frame = pd.DataFrame({"txt": lst})
    actual = frame["txt"].str.extract(pattern)
    assert actual[0].tolist() == lst



from io import StringIO
from typing import Dict, Sequence, Tuple

_OPTION: Tuple[str, Dict, bool] = ("?", {}, False)
_OPEN_PARENTHESIS: Tuple[str, Dict, bool] = ("(?:", {}, False)
_CLOSE_PARENTHESIS: Tuple[str, Dict, bool] = (")", {}, False)
_OPEN_BRACKETS: Tuple[str, Dict, bool] = ("[", {}, False)
_CLOSE_BRACKETS: Tuple[str, Dict, bool] = ("]", {}, False)
_ALTERNATION: Tuple[str, Dict, bool] = ("|", {}, False)


class _Trie:
    def __init__(self, words, left=r"\b", right=r"\b"):

        self.left = left
        self.right = right
        self.root = ({}, False)
        for word in sorted(words):
            node, is_terminal = self.root
            size = len(word)
            for i, char in enumerate(word):
                is_terminal = i == size - 1
                if char not in node:
                    node[char] = ({}, is_terminal)
                node, is_terminal = node[char]

    def _to_regex(self):

        stack = [(self.left,) + self.root]
        cumulative = StringIO()

        while stack:
            char, children, end = stack.pop()
            cumulative.write(char)

            if not children:
                continue  # skip

            if end:
                stack.append(_OPTION)

            single, multiple = [], []
            for key, (values, terminal) in children.items():
                if values:
                    multiple.append((key, values, terminal))
                else:
                    single.append((key, values, terminal))

            requires_character_set = len(single) > 1
            requires_alternation = (multiple and single) or len(multiple) > 1
            requires_group = requires_alternation or (end and multiple)

            if requires_group:
                stack.append(_CLOSE_PARENTHESIS)

            if requires_character_set:
                stack.append(_CLOSE_BRACKETS)

            for child in single:
                stack.append(child)

            if requires_character_set:
                stack.append(_OPEN_BRACKETS)

            if requires_alternation and single:
                stack.append(_ALTERNATION)

            if multiple:
                head, *tail = multiple
                stack.append(head)

                for child in tail:
                    stack.append(_ALTERNATION)
                    stack.append(child)

            if requires_group:
                stack.append(_OPEN_PARENTHESIS)

        return cumulative.getvalue()

    def make(self):
        return f"{self._to_regex()}{self.right}"


def make(words: Sequence[str], prefix: str = r"\b", suffix: str = r"\b"):
    """
    Create a string that represents a regular expression object from a set of strings

    Parameters
    ----------
    words : Sequence[str]
        Sequence or set of strings to be made into a regex

    prefix : str, optional
           Left delimiter for pattern

    suffix : str, optional
            Right delimiter for pattern

    Returns
    -------
    String
            A string representing a regular expression pattern

    Examples
    --------
    >>> import re
    >>> import trrex as tx
    >>> pattern = tx.make(["baby", "bat", "bad"])
    >>> re.findall(pattern, "The baby was scared by the bad bat.")
    ['baby', 'bad', 'bat']
    """

    return _Trie(words, left=prefix, right=suffix).make()



import re
from string import ascii_letters, punctuation

from hypothesis import example, given
from hypothesis.strategies import lists, text

from trrex import make


def compile(lst, flags=0, left=r"\b", right=r"\b"):
    return re.compile(make(lst, left, right), flags)


def test_findall():
    words = ["baby", "bad", "bat"]
    pattern = compile(words)
    assert pattern.findall("The baby was bitten by the bad bat") == words


def test_findall_ignore_case():
    words = ["baby", "bad", "bat"]
    pattern = compile(words, flags=re.IGNORECASE)
    actual = pattern.findall("The BABY was bitten by the bAD Bat")
    assert words == [w.lower() for w in actual]


def test_findall_whitespace():
    words = ["Ray Steam"]
    pattern = compile(words)
    actual = pattern.findall("Ray Steam is Steamboy")
    assert words == actual


def test_match():
    name = "Ray Steam"
    pattern = compile([name, "Edward Elric"])
    match = pattern.match(name)
    assert match is not None


def test_sub():
    replacements = {"Ray Steam": "Steamboy"}
    pattern = compile(list(replacements.keys()))

    def replace(match):
        return replacements.get(match.group(), "")

    actual = pattern.sub(replace, "The kid Ray Steam save the day")
    assert "The kid Steamboy save the day" == actual


def test_findall_punctuation():
    words = ["bab.y", "b#ad", "b?at"]
    pattern = compile(tuple(map(re.escape, word)) for word in words)
    assert pattern.findall("The bab.y was bitten by the b#ad b?at") == words


def test_match_punctuation():
    word = ":"
    pattern = compile([word], left="", right="")
    match = pattern.match(word)
    assert match is not None


def test_findall_emoticon():
    emoticons = [":)", ":D", ":("]
    pattern = compile(emoticons, left=r"(?<!\w)", right=r"(?!\w)")
    assert pattern.findall("The smile :), and the laugh :D and the sad :(") == emoticons


@given(text(alphabet=ascii_letters, min_size=1))
def test_single_string_match(s):
    pattern = compile([s])
    assert pattern.match(s) is not None


@example(lst=["B", "BA", "B"])
@given(lists(text(alphabet=ascii_letters, min_size=1)))
def test_multiple_string_match(lst):
    pattern = compile(lst)
    for word in lst:
        assert pattern.match(word) is not None


@given(lists(text(alphabet=ascii_letters + punctuation, min_size=1)))
def test_multiple_string_match_punctuation(lst):
    words = [tuple(map(re.escape, word)) for word in lst]
    pattern = compile(words, left="", right="")
    for word in lst:
        assert pattern.match(word) is not None



import re

from trrex import make


def compile(lst, flags=0, left=r"\b", right=r"\b"):
    return re.compile(make(lst, left, right), flags)


def test_left_pattern():
    pattern = compile(["bad", "bat", "bebop"], left="^")
    match = pattern.match("bad")
    assert match is not None


def test_left_pattern_no_match():
    pattern = compile(["bad", "bat", "bebop"], left="^")
    match = pattern.match(" bad")
    assert match is None


def test_right_pattern():
    pattern = compile(["bad", "bat", "bebop"], right="$")
    match = pattern.match("bad")
    assert match is not None


def test_right_pattern_no_match():
    pattern = compile(["bad", "bat", "bebop"], right="$")
    match = pattern.match("bad ")
    assert match is None


def test_prefix():
    pattern = compile(["bad", "bat", "bebop"], right=r"\w+")
    match = pattern.match("baddest bat in the planet")
    assert match is not None
    assert match.group() == "baddest"


def test_suffix():
    pattern = compile(["bebop"], left=r"\w+")
    match = pattern.search("cowboy bebebop")
    assert match is not None
    assert match.group() == "bebebop"
